# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import asyncio
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager
from typing import Callable, Dict, List
from zipfile import ZipFile

import boto3
import botocore.exceptions
import cloudpickle as pickle
from boto3.session import Session
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent_aws_plugins import AWSExecutor

from .scripts import PYTHON_EXEC_SCRIPT

app_log = logger.app_log
log_stack_info = logger.log_stack_info

executor_plugin_name = "AWSLambdaExecutor"

_EXECUTOR_PLUGIN_DEFAULTS = {
    "credentials_file": os.environ.get("AWS_SHARED_CREDENTIALS_FILE")
    or os.path.join(os.environ.get("HOME"), ".aws/credentials"),
    "profile": os.environ.get("AWS_PROFILE") or "default",
    "region": os.environ.get("AWS_REGION") or "us-east-1",
    "s3_bucket_name": "covalent-lambda-job-resources",
    "execution_role": "CovalentLambdaExecutionRole",
    "poll_freq": 5,
    "timeout": 60,
    "memory_size": 512,
    "cleanup": False,
}

LAMBDA_FUNCTION_NAME = "lambda-{dispatch_id}-{node_id}"
FUNC_FILENAME = "func-{dispatch_id}-{node_id}.pkl"
RESULT_FILENAME = "result-{dispatch_id}-{node_id}.pkl"
LAMBDA_DEPLOYMENT_ARCHIVE_NAME = "archive-{dispatch_id}-{node_id}.zip"
LAMBDA_FUNCTION_SCRIPT_NAME = "lambda_function.py"


class DeploymentPackageBuilder:
    """AWS Lambda deployment package (zip archive) builder

    Args:
        directory: Path to local filesystem on where to store the archive
        archive_name: Name of the deployment archive
        s3_bucket_name: Name of the AWS S3 bucket to be used to cache the deployment package
    """

    def __init__(self, directory: str, archive_name: str, s3_bucket_name: str):
        self.directory = directory
        self.archive_name = archive_name
        self.s3_bucket_name = s3_bucket_name
        self.target_dir = os.path.join(self.directory, "targets")
        self.deployment_archive = os.path.join(self.directory, self.archive_name)

    def install(self, pkg_name: str, pre: bool = False):
        """Install the necessary Python packages into the specified target directory

        Args:
            pkg_name: Name of the Python package to be installed
            pre: Boolean flag representing whether to install a pre-release version of the package

        Returns:
            None
        """
        if pre:
            return subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--target",
                    self.target_dir,
                    pkg_name,
                    "--pre",
                ],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT,
            )

        return subprocess.run(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--target",
                self.target_dir,
                "--upgrade",
                pkg_name,
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )

    def __enter__(self):
        """Create the zip archive"""
        if os.path.exists(self.target_dir):
            shutil.rmtree(self.target_dir)
            os.mkdir(self.target_dir)

        # Install the required python dependencies
        self.install("boto3")
        self.install("covalent==0.177.0")

        # Create zip archive with dependencies
        with ZipFile(self.deployment_archive, mode="w") as archive:
            for file_path in pathlib.Path(self.target_dir).rglob("*"):
                archive.write(
                    file_path, arcname=file_path.relative_to(pathlib.Path(self.target_dir))
                )

        return self.deployment_archive

    def __exit__(self, exc_type, exc_value, exc_tb):
        """None"""
        pass


class AWSLambdaExecutor(AWSExecutor):
    """AWS Lambda executor plugin

    Args:
        credentials_file: Path to AWS credentials file (default: `~/.aws/credentials`)
        profile: AWS profile (default: `default`)
        region: AWS region (default: `us-east-1`)
        s3_bucket_name: Name of a AWS S3 bucket that the executor can use to store temporary files (default: `covalent-lambda-job-resources`)
        execution_role: AWS IAM role name use to provision the Lambda function (default: `CovalentLambdaExecutionRole`)
        poll_freq: Time interval between successive polls to the lambda function (default: `5`)
        timeout: Duration in seconds before the Lambda function times out (default: `60`)
        cleanup: Flag represents whether or not to cleanup temporary files generated during execution (default: `True`)
    """

    def __init__(
        self,
        credentials_file: str,
        profile: str,
        region: str,
        s3_bucket_name: str,
        execution_role: str,
        poll_freq: int = None,
        timeout: int = None,
        memory_size: int = None,
        cleanup: bool = False,
    ) -> None:

        # AWSExecutor parameters
        required_attrs = {
            "credentials_file": credentials_file,
            "profile": profile,
            "region": region,
            "s3_bucket_name": s3_bucket_name,
            "execution_role": execution_role,
        }

        super().__init__(**required_attrs)

        # Lambda executor parameters
        self.poll_freq = poll_freq or get_config("executors.awslambda.poll_freq")
        self.timeout = timeout or get_config("executors.awslambda.timeout")
        self.memory_size = memory_size or get_config("executors.awslambda.memory_size")
        self.cleanup = cleanup

        self._cwd = tempfile.mkdtemp()
        self._key_exists = False

    @contextmanager
    def get_session(self) -> Session:
        """Yield a boto3 session to be used for instantiating AWS service clients/resources

        Args:
            None

        Returns:
            session: AWS boto3.Session object
        """
        yield boto3.Session(**self.boto_session_options())

    def _upload_task(self, workdir: str, func_filename: str):
        """
        Upload the function file to remote

        Args:
            workdir: Work dir on remote to upload file to
            func_filename: Name of the function file

        Returns:
            None
        """

        app_log.debug(f"Uploading function to S3 bucket {self.s3_bucket_name}")
        with self.get_session() as session:
            client = session.client("s3")
            try:
                with open(os.path.join(workdir, func_filename), "rb") as f:
                    client.upload_fileobj(f, self.s3_bucket_name, func_filename)
            except botocore.exceptions.ClientError as ce:
                app_log.exception(ce)
                raise
        app_log.debug(f"Function {func_filename} uploaded to S3 bucket {self.s3_bucket_name}")

    async def _is_lambda_active(self, function_name: str) -> bool:
        """Check if the lambda function is active of not

        Args:
            function_name: Name of the lambda function

        Returns:
            bool: Boolean value either True/False
        """
        # Check if lambda is active
        is_active = False
        with self.get_session() as session:
            lambda_client = session.client("lambda")
            while not is_active:
                lambda_state = lambda_client.get_function(FunctionName=function_name)
                app_log.debug(
                    f"Lambda function {function_name} state: {lambda_state['Configuration']['State']}"
                )
                if lambda_state["Configuration"]["State"] == "Active":
                    is_active = True
                else:
                    await asyncio.sleep(0.5)
                    continue
        return is_active

    async def _create_lambda(self, function_name: str, deployment_archive_name: str) -> Dict:
        """Create the AWS Lambda function

        Args:
            function_name: AWS Lambda function name
            deployment_archive_name: Lambda deployment zip archive name

        Returns:
            response: AWS boto3 client create_lambda response
        """
        with self.get_session() as session:
            iam_client = session.client("iam")
            role_arn = None
            try:
                response = iam_client.get_role(RoleName=f"{self.execution_role}")
                role_arn = response["Role"]["Arn"]
            except botocore.exceptions.ClientError as ce:
                app_log.exception(ce)
                raise

        lambda_create_kwargs = {
            "FunctionName": f"{function_name}",
            "Runtime": "python3.8",
            "Role": f"{role_arn}",
            "Handler": "lambda_function.lambda_handler",
            "Code": {"S3Bucket": f"{self.s3_bucket_name}", "S3Key": deployment_archive_name},
            "PackageType": "Zip",
            "Publish": True,
            "Timeout": self.timeout,
            "MemorySize": self.memory_size,
        }

        with self.get_session() as session:
            lambda_client = session.client("lambda")
            try:
                response = lambda_client.create_function(**lambda_create_kwargs)
                app_log.debug(f"Lambda function: {function_name} created: {response}")
            except botocore.exceptions.ClientError as ce:
                app_log.exception(ce)
                raise

        # Check if lambda is active
        is_active = await self._is_lambda_active(function_name)
        return "Active" if is_active else None

    def submit_task(self, function_name: str) -> Dict:
        """
        Submit the task by invoking the AWS Lambda function

        Args:
            function_name: AWS Lambda function name

        Returns:
            response: AWS boto3 client invoke lambda response
        """

        app_log.debug(f"Invoking AWS Lambda function {function_name}")

        with self.get_session() as session:
            client = session.client("lambda")
            try:
                return client.invoke(FunctionName=function_name)
            except botocore.exceptions.ClientError as ce:
                app_log.exception(ce)
                raise

    async def get_status(self, object_key: str):
        """
        Return status of availability of result object on remote machine

        Args:
            object_key: Name of the S3 object

        Returns:
            bool indicating whether the object exists or not on S3 bucket
        """

        with self.get_session() as session:
            while not self._key_exists:
                s3_client = session.client("s3")
                try:
                    current_keys = [
                        item["Key"]
                        for item in s3_client.list_objects(Bucket=self.s3_bucket_name)["Contents"]
                    ]
                    self._key_exists = object_key in current_keys

                    if not self._key_exists:
                        await asyncio.sleep(0.5)

                    return self._key_exists
                except botocore.exceptions.ClientError as ce:
                    app_log.exception(ce)
                    raise

        return self._key_exists

    async def _poll_task(self, object_key: str):
        """
        Poll task until its result is ready

        Args:
            object_key: Name of the object to check if present in S3
        """

        while not await self.get_status(object_key):
            app_log.debug(f"Polling object: {object_key}")
            await asyncio.sleep(self.poll_freq)

    def query_result(self, workdir: str, result_filename: str):
        """
        Fetch the result object from the S3 bucket

        Args:
            workdir: Path on the local file system where the pickled object is downloaded

        Returns:
            None
        """

        if self._key_exists:
            with self.get_session() as session:
                s3_client = session.client("s3")
                # Download file
                try:
                    s3_client.download_file(
                        self.s3_bucket_name,
                        result_filename,
                        os.path.join(workdir, result_filename),
                    )
                except botocore.exceptions.ClientError as ce:
                    app_log.exception(ce)
                    raise

        with open(os.path.join(workdir, result_filename), "rb") as f:
            result_object = pickle.load(f)

        return result_object

    async def setup(self, task_metadata: Dict):
        """AWS Lambda specific setup tasks

        Args:
            task_metadata: Dictionary containing the task dispatch_id and node_id

        Returns:
            None
        """
        dispatch_id = task_metadata["dispatch_id"]
        node_id = task_metadata["node_id"]
        workdir = os.path.join(self._cwd, dispatch_id)
        deployment_archive_name = LAMBDA_DEPLOYMENT_ARCHIVE_NAME.format(
            dispatch_id=dispatch_id, node_id=node_id
        )
        lambda_function_name = LAMBDA_FUNCTION_NAME.format(
            dispatch_id=dispatch_id, node_id=node_id
        )

        app_log.debug(f"Starting setup for task - {dispatch_id}-{node_id} ... ")

        if not os.path.exists(workdir):
            os.mkdir(workdir)

        app_log.debug(
            f"Creating the Lambda deployment archive at {os.path.join(workdir, deployment_archive_name)} ..."
        )
        with DeploymentPackageBuilder(
            workdir, deployment_archive_name, self.s3_bucket_name
        ) as deployment_archive:
            exec_script = PYTHON_EXEC_SCRIPT.format(
                func_filename=FUNC_FILENAME.format(dispatch_id=dispatch_id, node_id=node_id),
                result_filename=RESULT_FILENAME.format(dispatch_id=dispatch_id, node_id=node_id),
                s3_bucket_name=self.s3_bucket_name,
            )
            with open(os.path.join(workdir, LAMBDA_FUNCTION_SCRIPT_NAME), "w") as f:
                f.write(exec_script)
            # Add script to the deployment archive
            with ZipFile(deployment_archive, mode="a") as archive:
                archive.write(
                    os.path.join(workdir, LAMBDA_FUNCTION_SCRIPT_NAME),
                    arcname=LAMBDA_FUNCTION_SCRIPT_NAME,
                )

        app_log.debug(
            f"Lambda deployment archive: {os.path.join(workdir, deployment_archive_name)} created. Uploading to S3 ..."
        )

        # Upload archive to s3 bucket
        with self.get_session() as session:
            client = session.client("s3")
            try:
                client.upload_file(
                    deployment_archive, self.s3_bucket_name, deployment_archive_name
                )
            except botocore.exceptions.ClientError as ce:
                app_log.exception(ce)
                raise

        app_log.debug(f"Lambda deployment archive: {deployment_archive_name} uploaded to S3 ... ")

        # Create the lambda function
        app_log.debug(f"Creating AWS Lambda function {lambda_function_name} ...")
        state = await self._create_lambda(lambda_function_name, deployment_archive_name)
        app_log.debug(f"Lambda function: {lambda_function_name} created in state: {state}")

        app_log.debug(f"Finished setup for task - {dispatch_id}-{node_id} ... ")

    async def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict):
        """Run the executor

        Args:
            function: Python callable to be executed on the remote executor
            args: List of positional arguments to be passed to the function
            kwargs: Keyword arguments to be passed into the function
            task_metadata: Dictionary containing the task dispatch_id and node_id

        Returns:
            None
        """
        dispatch_id = task_metadata["dispatch_id"]
        node_id = task_metadata["node_id"]
        workdir = os.path.join(self._cwd, dispatch_id)

        func_filename = FUNC_FILENAME.format(dispatch_id=dispatch_id, node_id=node_id)
        result_filename = RESULT_FILENAME.format(dispatch_id=dispatch_id, node_id=node_id)
        lambda_function_name = LAMBDA_FUNCTION_NAME.format(
            dispatch_id=dispatch_id, node_id=node_id
        )

        app_log.debug(f"In run for task - {dispatch_id} - {node_id} ... ")

        app_log.debug("Pickling function, args and kwargs..")
        with open(os.path.join(workdir, func_filename), "wb") as f:
            pickle.dump((function, args, kwargs), f)

        # Upload pickled file to s3 bucket created
        self._upload_task(workdir, func_filename)

        # Invoke the created lambda
        lambda_invocation_response = self.submit_task(lambda_function_name)
        app_log.debug(f"Lambda function response: {lambda_invocation_response}")

        # Poll task
        await self._poll_task(result_filename)

        # Download the result object
        app_log.debug(f"Retrieving result for task - {dispatch_id} - {node_id}")
        result_object = self.query_result(workdir, result_filename)
        app_log.debug(f"Result retrived for task - {dispatch_id} - {node_id}")

        return result_object

    def cancel(self) -> None:
        """
        Cancel execution
        """
        raise NotImplementedError("Cancellation is currently not supported")

    async def teardown(self, task_metadata: Dict):
        """Cleanup temporary files and the Lambda function

        Args:
            task_metadata: Dictionary containing the task dispatch_id and node_id

        Returns:
            None
        """
        dispatch_id = task_metadata["dispatch_id"]
        node_id = task_metadata["node_id"]
        lambda_function_name = LAMBDA_FUNCTION_NAME.format(
            dispatch_id=dispatch_id, node_id=node_id
        )
        workdir = os.path.join(self._cwd, dispatch_id)

        app_log.debug(f"In teardown for task - {dispatch_id} - {node_id}")

        if self.cleanup:
            with self.get_session() as session:

                app_log.debug(f"Cleaning up resources created in S3 bucket {self.s3_bucket_name}")
                s3_resource = session.resource("s3")

                try:
                    s3_resource.Object(
                        self.s3_bucket_name,
                        FUNC_FILENAME.format(dispatch_id=dispatch_id, node_id=node_id),
                    ).delete()

                    s3_resource.Object(
                        self.s3_bucket_name,
                        RESULT_FILENAME.format(dispatch_id=dispatch_id, node_id=node_id),
                    ).delete()

                    s3_resource.Object(
                        self.s3_bucket_name,
                        LAMBDA_DEPLOYMENT_ARCHIVE_NAME.format(
                            dispatch_id=dispatch_id, node_id=node_id
                        ),
                    ).delete()

                    app_log.debug(f"All objects from bucket {self.s3_bucket_name} deleted")

                except botocore.exceptions.ClientError as ce:
                    app_log.exception(ce)
                    raise

                app_log.debug("Cleanup of resources in S3 bucket finished")

                app_log.debug(f"Deleting lambda function {lambda_function_name}")
                lambda_client = session.client("lambda")

                try:
                    response = lambda_client.delete_function(FunctionName=lambda_function_name)
                    app_log.debug(f"Lambda cleanup response: {response}")
                except botocore.exceptions.ClientError as ce:
                    app_log.exception(ce)
                    raise

                app_log.debug(f"Lambda function {lambda_function_name} deleted")

                app_log.debug(f"Cleaning up working directory {workdir}")
                if os.path.exists(workdir):
                    shutil.rmtree(workdir)
                app_log.debug(f"Working directory {workdir} deleted")

        app_log.debug(f"Finished teardown for task - {dispatch_id} - {node_id}")



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Tests for Covalent AWSLambda executor"""

import os

import botocore.exceptions
import pytest
from mock import AsyncMock, MagicMock, call

from covalent_awslambda_plugin import AWSLambdaExecutor, DeploymentPackageBuilder


@pytest.fixture
def lambda_executor():
    return AWSLambdaExecutor(
        credentials_file="~/.aws/credentials",
        profile="test_profile",
        region="us-east-1",
        s3_bucket_name="test_bucket_name",
        execution_role="test_lambda_role",
        poll_freq=30,
        timeout=10,
        memory_size=512,
        cleanup=True,
    )


def test_init():
    awslambda = AWSLambdaExecutor(
        credentials_file="~/.aws/credentials",
        profile="test_profile",
        region="us-east-1",
        execution_role="test_lambda_role",
        s3_bucket_name="test_bucket_name",
        poll_freq=30,
        timeout=10,
        memory_size=512,
        cleanup=True,
    )

    assert awslambda.credentials_file == "~/.aws/credentials"
    assert awslambda.profile == "test_profile"
    assert awslambda.region == "us-east-1"
    assert awslambda.execution_role == "test_lambda_role"
    assert awslambda.s3_bucket_name == "test_bucket_name"
    assert awslambda.poll_freq == 30
    assert awslambda.timeout == 10
    assert awslambda.memory_size == 512
    assert awslambda.cleanup


@pytest.mark.asyncio
async def test_setup_and_teardown_are_invoked(lambda_executor, mocker):
    "Simply assert that the setup, run and teardown methods are invoked when execute is called"
    lambda_executor.get_session = MagicMock()
    lambda_executor._is_lambda_active = MagicMock()
    lambda_executor._create_lambda = MagicMock()
    lambda_executor.submit_task = MagicMock()
    lambda_executor.get_status = MagicMock()
    lambda_executor.query_result = MagicMock()
    lambda_executor.setup = AsyncMock()
    lambda_executor.run = AsyncMock()
    lambda_executor.teardown = AsyncMock()

    await lambda_executor.execute(MagicMock(), MagicMock(), MagicMock(), MagicMock(), MagicMock())

    lambda_executor.setup.assert_awaited_once()
    lambda_executor.run.assert_awaited_once()
    lambda_executor.teardown.assert_awaited_once()


@pytest.mark.asyncio
async def test_setup_workdir_create(lambda_executor, mocker):
    dispatch_id = "aabbcc"
    node_id = 0
    task_metadata = {"dispatch_id": dispatch_id, "node_id": node_id}

    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    lambda_executor._create_lambda = AsyncMock()
    lambda_executor._cwd = "testdir"

    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")
    os_path_join_mock = mocker.patch("covalent_awslambda_plugin.awslambda.os.path.join")
    os_path_exists_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.os.path.exists", return_value=False
    )
    os_mkdir_mock = mocker.patch("covalent_awslambda_plugin.awslambda.os.mkdir")
    mocker.patch("covalent_awslambda_plugin.awslambda.open")
    mocker.patch("covalent_awslambda_plugin.awslambda.ZipFile")
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder", return_value=MagicMock()
    )

    await lambda_executor.setup(task_metadata)

    os_path_exists_mock.assert_called_once()
    os_mkdir_mock.assert_called_once()


@pytest.mark.asyncio
async def test_deployment_package_builder(lambda_executor, mocker):
    """Test if the deployment package builder context manager enter method is invoked"""
    dispatch_id = "aabbcc"
    node_id = 0

    task_metadata = {"dispatch_id": dispatch_id, "node_id": node_id}

    deployment_package_builder_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder", return_value=MagicMock()
    )
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )

    lambda_executor._create_lambda = AsyncMock()
    mocker.patch("covalent_awslambda_plugin.awslambda.app_log")
    mocker.patch("covalent_awslambda_plugin.awslambda.ZipFile")
    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True)
    mocker.patch("covalent_awslambda_plugin.awslambda.os.mkdir")
    mocker.patch("covalent_awslambda_plugin.awslambda.open")

    await lambda_executor.setup(task_metadata)

    deployment_package_builder_mock.return_value.__enter__.assert_called_once()
    deployment_package_builder_mock.return_value.__exit__.assert_called_once()


def test_deployment_package_builder_base_setup(mocker):
    workdir = "testdir"
    archive_name = "test_archive"
    s3_bucket_name = "test_bucket"

    zipfile_mock = mocker.patch("covalent_awslambda_plugin.awslambda.ZipFile")
    os_path_exists_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True
    )
    shutil_rmtree_mock = mocker.patch("covalent_awslambda_plugin.awslambda.shutil.rmtree")
    os_mkdir_mock = mocker.patch("covalent_awslambda_plugin.awslambda.os.mkdir")
    install_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.install"
    )

    pkg_bldr = DeploymentPackageBuilder(workdir, archive_name, s3_bucket_name)
    with DeploymentPackageBuilder(workdir, archive_name, s3_bucket_name) as bldr:
        pass

    os_path_exists_mock.assert_called_once_with(pkg_bldr.target_dir)
    shutil_rmtree_mock.assert_called_once_with(pkg_bldr.target_dir)
    os_mkdir_mock.assert_called_once_with(pkg_bldr.target_dir)
    assert install_mock.call_count == 2
    zipfile_mock.assert_called_once_with(pkg_bldr.deployment_archive, mode="w")


def test_deployment_package_builder_install_method(mocker):
    workdir = "testdir"
    archive_name = "test_archive"
    s3_bucket_name = "test_bucket"

    subprocess_mock = mocker.patch("covalent_awslambda_plugin.awslambda.subprocess.run")
    mocker.patch("covalent_awslambda_plugin.awslambda.ZipFile")
    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True)
    mocker.patch("covalent_awslambda_plugin.awslambda.shutil.rmtree")
    mocker.patch("covalent_awslambda_plugin.awslambda.os.mkdir")

    with DeploymentPackageBuilder(workdir, archive_name, s3_bucket_name) as bldr:
        pass

    assert subprocess_mock.call_count == 2


@pytest.mark.asyncio
async def test_function_pickle_dump(lambda_executor, mocker):
    def f(x):
        return x

    lambda_executor._upload_task = MagicMock()
    lambda_executor.submit_task = MagicMock()
    lambda_executor._poll_task = AsyncMock()
    lambda_executor.get_session = MagicMock()

    lambda_executor.query_result = MagicMock()

    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.join")
    file_open_mock = mocker.patch("covalent_awslambda_plugin.awslambda.open")
    pickle_dump_mock = mocker.patch("covalent_awslambda_plugin.awslambda.pickle.dump")

    await lambda_executor.run(f, 1, {}, {"dispatch_id": "aabbcc", "node_id": 0})

    file_open_mock.return_value.__enter__.assert_called()
    pickle_dump_mock.assert_called_once()


def test_upload_fileobj(lambda_executor, mocker):

    lambda_executor.get_session = MagicMock()

    mocker.patch("covalent_awslambda_plugin.awslambda.app_log")
    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.join")
    file_open_mock = mocker.patch("covalent_awslambda_plugin.awslambda.open")
    mocker.patch("covalent_awslambda_plugin.awslambda.pickle.dump")

    lambda_executor._upload_task("test_workdir", "test_func_filename")

    file_open_mock.assert_called_once()

    lambda_executor.get_session.assert_called_once()
    lambda_executor.get_session.return_value.__enter__.assert_called_once()
    lambda_executor.get_session.return_value.__enter__.return_value.client.assert_called_once_with(
        "s3"
    )
    file_open_mock.assert_called()
    lambda_executor.get_session.return_value.__enter__.return_value.client.return_value.upload_fileobj.assert_called_once()


def test_upload_fileobj_exception(lambda_executor, mocker):
    def f(x):
        return x

    lambda_executor.get_session = MagicMock()
    lambda_executor._create_lambda = MagicMock()
    lambda_executor.submit_task = MagicMock()
    lambda_executor.query_result = MagicMock()
    lambda_executor._cleanup = MagicMock()

    client_error_mock = botocore.exceptions.ClientError(MagicMock(), MagicMock())
    lambda_executor.get_session.return_value.__enter__.return_value.client.return_value.upload_fileobj.side_effect = (
        client_error_mock
    )

    mocker.patch("covalent_awslambda_plugin.awslambda.open")
    mocker.patch("covalent_awslambda_plugin.awslambda.pickle.dump")
    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    with pytest.raises(botocore.exceptions.ClientError):
        lambda_executor._upload_task("test_workdir", "test_func_filename")
        app_log_mock.exception.assert_called_with(client_error_mock)


@pytest.mark.asyncio
async def test_setup(lambda_executor, mocker):
    dispatch_id = "aabbcc"
    node_id = 0
    target_metdata = {"dispatch_id": dispatch_id, "node_id": node_id}

    lambda_executor.get_session = MagicMock()
    lambda_executor._create_lambda = AsyncMock()

    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")
    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True)
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.__enter__", return_value=None
    )
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.__exit__", return_value=None
    )
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.install", return_value=None
    )
    mocker.patch("covalent_awslambda_plugin.awslambda.ZipFile")
    mocker.patch("covalent_awslambda_plugin.awslambda.open")

    await lambda_executor.setup(target_metdata)

    lambda_executor.get_session.assert_called()
    lambda_executor.get_session.return_value.__enter__.assert_called()
    lambda_executor.get_session.return_value.__enter__.return_value.client.assert_called_with("s3")
    lambda_executor.get_session.return_value.__enter__.return_value.client.return_value.upload_file.assert_called()


@pytest.mark.asyncio
async def test_setup_exception(lambda_executor, mocker):
    dispatch_id = "aabbcc"
    node_id = 0
    task_metdata = {"dispatch_id": dispatch_id, "node_id": node_id}

    lambda_executor.get_session = MagicMock()
    lambda_executor._create_lambda = MagicMock()

    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True)
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.__enter__", return_value=None
    )
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.__exit__", return_value=None
    )
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.install", return_value=None
    )
    mocker.patch("covalent_awslambda_plugin.awslambda.ZipFile")
    mocker.patch("covalent_awslambda_plugin.awslambda.open")

    client_error_mock = botocore.exceptions.ClientError(MagicMock(), MagicMock())
    lambda_executor.get_session.return_value.__enter__.return_value.client.return_value.upload_file.side_effect = (
        client_error_mock
    )

    with pytest.raises(botocore.exceptions.ClientError):
        await lambda_executor.setup(task_metdata)

        lambda_executor.get_session.assert_called()
        lambda_executor.get_session.return_value.__enter__.assert_called()
        lambda_executor.get_session.return_value.__enter__.return_value.client.assert_called_with(
            "s3"
        )
        app_log_mock.exception.assert_called_once()


@pytest.mark.asyncio
async def test_create_lambda_invocation(lambda_executor, mocker):
    """Test to see if create lambda method gets invoked"""
    dispatch_id = "aabbcc"
    node_id = 0
    task_metadata = {"dispatch_id": dispatch_id, "node_id": node_id}

    lambda_executor.get_session = MagicMock()
    lambda_executor._create_lambda = AsyncMock()
    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True)
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.__enter__", return_value=None
    )
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.__exit__", return_value=None
    )
    mocker.patch(
        "covalent_awslambda_plugin.awslambda.DeploymentPackageBuilder.install", return_value=None
    )
    mocker.patch("covalent_awslambda_plugin.awslambda.ZipFile")
    mocker.patch("covalent_awslambda_plugin.awslambda.open")

    await lambda_executor.setup(task_metadata)

    lambda_executor._create_lambda.assert_called_once_with(
        f"lambda-{dispatch_id}-{node_id}", f"archive-{dispatch_id}-{node_id}.zip"
    )


@pytest.mark.asyncio
async def test_create_lambda_iam_get_role(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    lambda_executor._is_lambda_active = AsyncMock(return_value=True)
    dispatch_id = "abcd"
    node_id = 0

    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"
    archive_name = f"archive-{dispatch_id}-{node_id}.zip"

    mocker.patch("covalent_awslambda_plugin.awslambda.app_log")
    mocker.patch("covalent_awslambda_plugin.awslambda.open")

    await lambda_executor._create_lambda(lambda_function_name, archive_name)

    session_mock.return_value.__enter__.assert_called()
    session_mock.return_value.__enter__.return_value.client.assert_has_calls([call("iam")])
    session_mock.return_value.__enter__.return_value.client.return_value.get_role.assert_called_with(
        RoleName=lambda_executor.execution_role
    )


@pytest.mark.asyncio
async def test_create_lambda_iam_get_role_exception(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    dispatch_id = "abcd"
    node_id = 0
    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"
    archive_name = f"archive-{dispatch_id}-{node_id}.zip"
    lambda_executor._is_lambda_active = AsyncMock(return_value=True)

    client_error_mock = botocore.exceptions.ClientError(MagicMock(), MagicMock())
    session_mock.return_value.__enter__.return_value.client.return_value.get_role.return_value = {
        "Role": {"Arn": "test_arn"}
    }

    session_mock.return_value.__enter__.return_value.client.return_value.get_role.side_effect = (
        client_error_mock
    )
    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    with pytest.raises(botocore.exceptions.ClientError):
        await lambda_executor._create_lambda(lambda_function_name, archive_name)
        app_log_mock.exception.assert_called_with(client_error_mock)

    session_mock.return_value.__enter__.return_value.client.return_value.get_role.assert_called()


@pytest.mark.asyncio
async def test_create_lambda_create_function(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    lambda_executor._is_lambda_active = AsyncMock(return_value=True)
    dispatch_id = "abcd"
    node_id = 0

    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"
    archive_name = f"archive-{dispatch_id}-{node_id}.zip"

    mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    await lambda_executor._create_lambda(lambda_function_name, archive_name)

    session_mock.return_value.__enter__.return_value.client.assert_has_calls([call("lambda")])
    session_mock.return_value.__enter__.return_value.client.return_value.create_function.assert_called()


@pytest.mark.asyncio
async def test_create_lambda_create_function_exception(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    lambda_executor._is_lambda_active = AsyncMock(return_value=True)
    dispatch_id = "abcd"
    node_id = 0

    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"
    archive_name = f"archive-{dispatch_id}-{node_id}.zip"

    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    client_error_mock = botocore.exceptions.ClientError(MagicMock(), MagicMock())
    session_mock.return_value.__enter__.return_value.client.return_value.create_function.side_effect = (
        client_error_mock
    )

    with pytest.raises(botocore.exceptions.ClientError):
        await lambda_executor._create_lambda(lambda_function_name, archive_name)
        app_log_mock.exception.assert_called_with(client_error_mock)


@pytest.mark.asyncio
async def test_is_lambda_active(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )

    dispatch_id = "abcd"
    node_id = 0
    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"

    session_mock.return_value.__enter__.return_value.client.return_value.get_function.return_value = {
        "Configuration": {"State": "Active"}
    }
    mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    await lambda_executor._is_lambda_active(lambda_function_name)

    session_mock.return_value.__enter__.return_value.client.return_value.get_function.assert_called_with(
        FunctionName=lambda_function_name
    )


def test_submit_task(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    dispatch_id = "abcd"
    node_id = 0
    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"

    lambda_executor.submit_task(lambda_function_name)

    session_mock.return_value.__enter__.return_value.client.assert_called_with("lambda")
    session_mock.return_value.__enter__.return_value.client.return_value.invoke.assert_called_with(
        FunctionName=lambda_function_name
    )


def test_submit_task_exception(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    dispatch_id = "abcd"
    node_id = 0
    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"

    client_error_mock = botocore.exceptions.ClientError(MagicMock(), MagicMock())
    session_mock.return_value.__enter__.return_value.client.return_value.invoke.side_effect = (
        client_error_mock
    )

    with pytest.raises(botocore.exceptions.ClientError):
        lambda_executor.submit_task(lambda_function_name)
        app_log_mock.exception.assert_called_with(client_error_mock)


@pytest.mark.asyncio
async def test_get_status(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )

    result_filename = "test_file"

    session_client_mock = session_mock.return_value.__enter__.return_value.client
    s3_client_list_objects_mock = (
        session_mock.return_value.__enter__.return_value.client.return_value.list_objects
    )
    s3_client_list_objects_mock.return_value = {"Contents": [{"Key": result_filename}]}
    mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    key_exists = await lambda_executor.get_status(result_filename)

    session_client_mock.assert_called_with("s3")
    s3_client_list_objects_mock.assert_called_with(Bucket=lambda_executor.s3_bucket_name)
    assert key_exists


@pytest.mark.asyncio
async def test_get_status_else_path(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )

    result_filename = "test_file"

    session_client_mock = session_mock.return_value.__enter__.return_value.client
    s3_client_list_objects_mock = (
        session_mock.return_value.__enter__.return_value.client.return_value.list_objects
    )
    s3_client_list_objects_mock.return_value = {"Contents": [{"Key": "not_test_file"}]}

    return_value = await lambda_executor.get_status(result_filename)

    session_client_mock.assert_called_with("s3")
    s3_client_list_objects_mock.assert_called_with(Bucket=lambda_executor.s3_bucket_name)
    assert not return_value


def test_query_result(lambda_executor, mocker):
    result_filename = "test_file"
    workdir = "test_dir"
    lambda_executor._key_exists = True

    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    session_client_mock = session_mock.return_value.__enter__.return_value.client
    s3_client_mock = session_client_mock.return_value.download_file
    open_mock = mocker.patch("covalent_awslambda_plugin.awslambda.open")
    pickle_load_mock = mocker.patch("covalent_awslambda_plugin.awslambda.pickle.load")

    lambda_executor.get_status = MagicMock(return_value=True)

    lambda_executor.query_result(workdir, result_filename)

    session_client_mock.assert_called_once_with("s3")
    s3_client_mock.assert_called_once_with(
        lambda_executor.s3_bucket_name, result_filename, os.path.join(workdir, result_filename)
    )
    open_mock.assert_called_once_with(os.path.join(workdir, result_filename), "rb")
    pickle_load_mock.assert_called_once_with(open_mock.return_value.__enter__.return_value)


def test_query_result_exception(lambda_executor, mocker):
    result_filename = "test_file"
    workdir = "test_dir"
    lambda_executor._key_exists = True

    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )

    session_client_mock = session_mock.return_value.__enter__.return_value.client
    s3_client_mock = session_client_mock.return_value.download_file
    client_error_mock = botocore.exceptions.ClientError(MagicMock(), MagicMock())
    s3_client_mock.side_effect = client_error_mock

    open_mock = mocker.patch("covalent_awslambda_plugin.awslambda.open")
    pickle_load_mock = mocker.patch("covalent_awslambda_plugin.awslambda.pickle.load")
    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")

    with pytest.raises(botocore.exceptions.ClientError):
        lambda_executor.query_result(workdir, result_filename)

        session_client_mock.assert_called_once_with("s3")

        s3_client_mock.assert_called_once_with(
            lambda_executor.s3_bucket_name, result_filename, os.path.join(workdir, result_filename)
        )

        app_log_mock.exception.assert_called_once_with(client_error_mock)
        open_mock.assert_called_once_with(os.path.join(workdir, result_filename), "rb")
        pickle_load_mock.assert_called_once_with(open_mock.return_value.__enter__.return_value)


@pytest.mark.asyncio
async def test_teardown(lambda_executor, mocker):
    dispatch_id = "aabbcc"
    node_id = 1
    task_metadata = {"dispatch_id": dispatch_id, "node_id": node_id}

    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"
    lambda_executor.cleanup = True
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )
    mocker.patch("covalent_awslambda_plugin.awslambda.app_log")
    mocker.patch("covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True)
    shutil_rmtree_mock = mocker.patch("covalent_awslambda_plugin.awslambda.shutil.rmtree")
    session_resource_mock = session_mock.return_value.__enter__.return_value.resource
    session_object_mock = session_resource_mock.return_value.Object
    object_delete_mock = session_object_mock.return_value.delete
    session_client_mock = session_mock.return_value.__enter__.return_value.client
    lambda_client_mock = session_client_mock.return_value

    await lambda_executor.teardown(task_metadata)

    session_resource_mock.assert_called_with("s3")
    session_object_mock.assert_has_calls(
        [call(lambda_executor.s3_bucket_name, f"func-{dispatch_id}-{node_id}.pkl")]
    )
    session_object_mock.assert_has_calls(
        [call(lambda_executor.s3_bucket_name, f"result-{dispatch_id}-{node_id}.pkl")]
    )
    session_object_mock.assert_has_calls(
        [call(lambda_executor.s3_bucket_name, f"archive-{dispatch_id}-{node_id}.zip")]
    )
    assert object_delete_mock.call_count == 3
    session_client_mock.assert_called_with("lambda")
    lambda_client_mock.delete_function.assert_called_once_with(FunctionName=lambda_function_name)
    shutil_rmtree_mock.assert_called_once()


@pytest.mark.asyncio
async def test_teardown_exception(lambda_executor, mocker):
    session_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.AWSLambdaExecutor.get_session",
        return_value=MagicMock(),
    )

    dispatch_id = "aabbcc"
    node_id = 1
    task_metadata = {"dispatch_id": dispatch_id, "node_id": node_id}

    workdir = "test_workdir"
    lambda_function_name = f"lambda-{dispatch_id}-{node_id}"

    app_log_mock = mocker.patch("covalent_awslambda_plugin.awslambda.app_log")
    exit_mock = mocker.patch("covalent_awslambda_plugin.awslambda.exit")

    s3_resource_mock = session_mock.return_value.__enter__.return_value.resource
    delete_object_mock = s3_resource_mock.return_value.Object.return_value.delete

    client_error_mock = botocore.exceptions.ClientError(MagicMock(), MagicMock())
    delete_object_mock.side_effect = client_error_mock

    client_mock = session_mock.return_value.__enter__.return_value.client
    lambda_client_mock = client_mock.return_value
    lambda_delete_function_mock = lambda_client_mock.delete_function
    lambda_delete_function_mock.side_effect = client_error_mock
    os_path_exists_mock = mocker.patch(
        "covalent_awslambda_plugin.awslambda.os.path.exists", return_value=True
    )
    shutil_rmtree_mock = mocker.patch("covalent_awslambda_plugin.awslambda.shutil.rmtree")

    with pytest.raises(botocore.exceptions.ClientError):
        await lambda_executor.teardown(task_metadata)
        s3_resource_mock.assert_called_once_with("s3")
        app_log_mock.exception.assert_called_with(client_error_mock)
        client_mock.assert_called_once_with("lambda")
        lambda_delete_function_mock.assert_called_with(FunctionName=lambda_function_name)
        assert app_log_mock.exception.call_count == 2
        assert exit_mock.call_count == 2
        os_path_exists_mock.assert_called_once()
        shutil_rmtree_mock.assert_called_once()
